import fs from "fs"
import path from "path"
import { fileURLToPath } from 'url';

import { IOS, System, Serial, Utils } from "syrus4-nodejs"

import dayjs from "../utils/dayjs.js"

import storage from "../services/storage.js"
import { Queue } from "../services/Queue.js"

import { getCounters } from "./TAIP.d/counters.js"
import { ECU_LIST, ECUToTAIP } from "./TAIP.d/ecu.js"
import { getIO, getAnalogs } from "./TAIP.d/io.js"
import { getSatcom, getIB_IS, getEA, getRI, getFF} from "./TAIP.d/peripherals.js"
import { pad, getLat, getLon, getSC, __noop } from "./TAIP.d/utils.js"
import { MediaManagerMixin } from "./TAIP.d/MediaManager.js"
import { BacklogMixin } from "./TAIP.d/Backlog.js";

const __filename = fileURLToPath(import.meta.url)
const __dirname = path.dirname(__filename)
const GPS_EPOCH = dayjs.unix(315964800)

/**
 * TAIP class manage all the communication in TAIP protocol and delivers order to syrusjs from TAIP commands
 */
class BaseTAIP {
	// events
	constructor({ name, client, ack, onCommand, autoDequeue = false }) {
		let that = this
		this.retry = null;
		this.retries = 0;
		this.name = name;
		this.queue = new Queue(`peg_queue_${name}`);
		this.queue.ready().then(list => {});
		this.connected = false;
		this.imei;
		this.IMSI;
		this.ICCID;
		this.carrier;
		this.trips_events = false;
		this.onCommand = onCommand || _noop;
		this.ack = ack || 'imei'
		this.ack_counter = 0
		this.client = client
		this.taip_buffer = ''
		this.taip_buffer_timeout = null
		this.__ready = null
		this.ready = new Promise((resolve)=>{ that.__ready = resolve })
		this.__flushing = false
		this.init();
	}

	// Preparing data for and connections callbacks
	async init() {
		let modem, info, version
		try {
			info = await System.info();
		} catch (error){
			console.error(`TCP:System.info()`, error);
			info = {}
		}
		try{
			modem = await System.modem();
		} catch (error) {
			console.error("TCP:System.modem()", error);
			modem = {};
		}
		// Getting the active sim
		this.imei = modem.IMEI;
		if (modem.SIM_STATE == "READY") {
			this.IMSI = modem.SIM_IMSI || modem.eSIM_IMSI || "";
			this.ICCID = modem.SIM_ID || modem.eSIM_ID || "";
		} else {
			this.IMSI = modem.eSIM_IMSI || modem.SIM_IMSI || "";
			this.ICCID = modem.eSIM_ID || modem.SIM_ID || "";
		}
		// Getting version and carrier
		version = info.apexVersion ? info.apexVersion.replace("apex-", "") : "00.00.00";
		version = version.split("-")[0] + ".00";
		this.carrier = modem.OPERATOR || "unknown";

		// ack
		this.ack_counter = await storage.get(`${this.name}_ack`);
		if (!this.ack_counter) this.ack_counter = 0

		// FTP
		this._photo_state = await storage.hgetall(`${this.name}_photo_state`)
		this.media_state = await storage.hgetall(`${this.name}_media_state`)
		this.media_state = {...this._photo_state, ...this.media_state}

		this.ftp = await storage.hgetall(`${this.name}_ftp_config`)
		console.log(`tcp(${this.name}):ftp`, this.ftp)

		// Establish the callbacks for the channel (TCP, HTTP, file, etc)
		this.client.onConnect = () => {
			this.connected = true;
			// some clients (i.e satcom, don't need a welcome message)
			if(this.client.disable_welcome_message) return this.flush()
			// Sending salute after connection
			let package_json = JSON.parse(fs.readFileSync(path.resolve(__dirname, "../package.json")))
			let additional = ''
			additional += `model=Syrus 4 LTE,syrusjs_version=${package_json.version}`
			additional += `,os_version=${info.apexVersion}`
			additional += `,kernel_version=${info.kernelVersion},platform=APEX OS`
			let signature = `ID=${this.imei}`
			if (this.ack == "disabled") signature = `SA=0,0;${signature}`
			let salut = `>RXART;${version};EG25S;${additional};${signature}<`
			this.priorityPush(salut)
		};

		// hook to channel's onMessage
		this.client.onMessage = data => {
			// continue buffer
			if (this.taip_buffer.length){
				clearTimeout(this.taip_buffer_timeout)
				data = `${this.taip_buffer}${data}`
				this.taip_buffer = ''
			}
			let buffer = data.split('>').map(i => i.split('<')).reduce((a, b) => a.concat(b))
			let messages = []
			buffer.forEach(message => {
				if (!message.trim().length) return
				if (message == this.imei) return messages.push(message)
				if (data.includes(`>${message}<`)) return messages.push(`>${message}<`)
				this.taip_buffer = `>${message}`
			})
			// console.log({taip_buffer: this.taip_buffer, messages})
			if (this.taip_buffer.length){
				clearTimeout(this.taip_buffer_timeout)
				this.taip_buffer_timeout = setTimeout(() => {
					console.log('taip:buffered', this.taip_buffer)
					this.responseHandler(this.taip_buffer)
					this.taip_buffer = ''
				}, 1000 * 3)
			}
			messages.map(m => this.responseHandler(m))
		};

		this.client.onDisconnect = () => {
			console.log("connection lost", this.name);
			this.connected = false;
		};

		this.client.keepAliveMessage = this.imei;
		this.client.connect();
		this.__ready()
	}

	/**
	 * default send command this is called from the main, process the event To the protocol
	 * @param {Object<any>} variables syrusjs_variables copy
	 * @param {Object<any>} opts event defintion parameters, min_duration, trigger rater,etc
	 * @param {Object<any>} fields fieldset from the event
	 * @param {string} label the label from the event
	 * @param {string} code the code from the event
	 */
	async send(data, opts, fields, label, code) {
		await this.ready
		this.fields = fields; // saving a copy from the fieldset to allow taip changes the counters when taip required
		if (!opts.ack) opts.ack = this.ack
		if ("off".indexOf(opts.ack) > -1) opts.ack = "disabled"
		if ("on,enabled".indexOf(opts.ack) > -1) opts.ack = "seq"
		// backlog switch
		if (opts.backlog) {
			const backlog_file = data[opts.backlog] || opts.backlog
			this.processBacklog(backlog_file, label, code)
			return
		}
		let taip_message = await this.JSONtoTAIP(data, opts, fields, label, code);
		this.push(taip_message);
	}

	/**
	 * adds messages to queue
	*/
	async push(message, priority=false) {
		if (priority) this.queue.unshift(message)
		else this.queue.push(message)
		if (this.queue.length && this.connected) await this.flush()
	}

	// Used for pushing at event at the top of queue
	// no too much usage right now useful for debug
	async priorityPush(message) {
		// await this.push(message, true)
		this.client.send(message);
	}

	// Called to dequeue an event its called after every ack or when the connection is established
	async flush(recursive) {
		await this.queue.ready()
		if (this.queue.length == 0) return this.__flushing = false
		if (this.__flushing && !recursive) return
		this.__flushing = true
		let message = this.queue[0]
		// discard invalid messages
		if (!message || message.includes("ID=undefined")){
			this.queue.shift()
			this.flush(true)
			return
		}
		// set retry
		clearTimeout(this.retry);
		this.retry = setTimeout(() => {
			console.log(`no:ack: ${++this.retries}`, message);
			if (this.client.no_disconnects || this.retries < 3) return this.flush(true)
			this.retries = 0
			this.client.disconnect();
			return
		}, 1000 * 30);
		// send message
		this.client.send(message, (error) => {
			if (error){
				console.error(error)
				this.client.disconnect()
				return
			}
			// keep processing queue if no ack is needed
			if (this.client.no_acks) return this.responseHandler(this.imei);
			if (message.includes(';SA=0,')) return this.responseHandler(this.imei);
		});
	}

	/**
	 * Manage all the TAIP commands received from the server
	 * @param {*} data
	 */
	responseHandler(data) {
		// ACK messages(imei or seq)
		if (data.trim().startsWith(this.imei) || data.includes(">SAK;")){
			clearTimeout(this.retry);
			this.retry = null;

			// no more events, invalid message
			let message = this.queue[0]
			if (this.queue.length == 0 || !message) return

			// ack checks
			// trip confirmation
			if (data.includes(">SAK;t") && !message.includes(">RXATN")) return
			let ack_required = message.includes(";SA=1,")
			if(ack_required){
				let ack_id = message.split(";SA=1,")[1].split(';')[0]
				if (!data.includes(`>SAK;1,${ack_id},`)) return this.flush(true)
			}
			// send photos after PS tag is acknowledged
			if (message.includes(";PS")) this.sendMedia()
			// pop event from queue
			this.retries = 0
			this.queue.shift()
			this.flush(true)
			return
		}
		const trimmed = data.split(";")[0].replace(">", "").replace("<", "")
		// Queries
		if (data.indexOf(">QXART") != -1) {
			System.info()
				.then(info => {
					let version = info.apexVersion ? info.apexVersion.replace("apex-", "") : "00.00.00";
					version = version.split("-")[0] + ".00";
					this.priorityPush(this.msgbuilder(this.RPV(`>RXART;${version};EG25S;`), data));
				})
				.catch(err => {
					console.error(err);
					let version = "00.00.00";
					this.priorityPush(this.msgbuilder(this.RPV(`>RXART;${version};EG25S;`), data));
				});
			return;
		}
		if (data.indexOf(">QPV") != -1) {
			this.priorityPush(this.msgbuilder(this.RPV(this.last_data), data));
			return;
		}
		if (data.indexOf(">QVR") != -1) {
			this.priorityPush(this.msgbuilder(`>RVR7.0.${1}`, data));
			return;
		}
		if (data.indexOf(">QXARS") != -1) {
			this.priorityPush(this.msgbuilder(`RXARS;${this.IMSI};${this.carrier};${this.ICCID.replace("+QCCID: ", "")}`, data));
			return;
		}
		if (data.indexOf(">QXAEM0") != -1) {
			// using await becomes painful upstream
			Utils.OSExecute("apx-ecu list_parameters").then((ecu_parameters) => {
				let xaem_groups = [
					new Array(64).fill(0),
					new Array(64).fill(0),
					new Array(64).fill(0),
				]
				for (const ecu_key of Object.keys(ecu_parameters)){
					let ecu_item = ECU_LIST[ecu_key]
					if (!ecu_item) continue
					let { $taip_xaem_group, $taip_xaem_bit } = ecu_item
					if (!$taip_xaem_group || !$taip_xaem_bit) continue
					xaem_groups[$taip_xaem_group - 1][$taip_xaem_bit - 1] = 1
				}
				let ecu_response = xaem_groups.map(group => parseInt(group.join(''), 2)).join(',')
				this.priorityPush(this.msgbuilder(`RXAEM0T,60,2,${ecu_response},6.1.3-apex`, data));
			})
			return;
		}
		if (data.indexOf(">QXATT") != -1) {
			if (this.trips_events) this.priorityPush(this.msgbuilder("RXATT010000", data));
			else this.priorityPush(this.msgbuilder("RXATTU", data));
			return;
		}
		if (data.indexOf(">QSSIP") != -1) {
			let output = parseInt(data[6]);
			IOS.getInputState(`IN${output}`).then(state => {
				this.priorityPush(this.msgbuilder(">RSSIP" + output + state, data));
			});
			return;
		}
		if (data.indexOf(">QSSXP") != -1) {
			output = parseInt(data[6]);
			IOS.getInputState(`OUT${output}`).then(state => {
				this.priorityPush(this.msgbuilder(">RSSXP" + output + state, data));
			});
			return;
		}
		if (data.indexOf(">QFP") != -1) {
			this.priorityPush(`>RFPA;ID=${this.imei}<`);
			return;
		}
		if (data.indexOf(">QXASMS") != -1) {
			Serial.getSerialModemState().then(state => {
				let connected = state.state == "connected" ? 1 : 0
				let is_satcom = state.type == "satcom"
				let msg = `RXASMS${connected}`
				state.msg_err = state.msg_err ? state.msg_err : ''
				if (is_satcom && connected){
					msg += `${state.signal}${state.id}${state.msg_err}`
				}
				this.priorityPush(this.msgbuilder(msg, data));
			}).catch((error)=>{
				console.error(error)
				this.priorityPush(this.msgbuilder("RXASMS0", data));
			});
			return
		}
		if (data.indexOf(">QXASF00") != -1) {
			this.priorityPush(
				this.msgbuilder(`RXASF00${this.queue.length}+[0F]M,0B,0%`, data)
			);
			return;
		}

		// Setters
		if (data.indexOf(">SSL") != -1) {
			this.onCommand({
				isSyrusLang: true,
				command: data
					.replace(">SSL", "")
					.split("<")[0]
					.split(";ID=")[0]
					.split(";SI=")[0]
			}).then(response => {
				if (!response) response = "OK";
				this.priorityPush(this.msgbuilder(
					`RSL${
						typeof response == "string"
							? response.substring(0, 512).replace(/\n/g, "")
							: JSON.stringify(response).substring(0, 512).replace(/\n/g, "")
					}`,
					data
				))
			}).catch(err => {
				let code = err.error ? err.error.code : "1";
				let errorText = JSON.stringify(err);
				if (err.errorText) {
					errorText = err.errorText;
				} else if (err.output) {
					errorText = err.output;
				}
				try {
					errorText = JSON.parse(errorText).error;
				} catch (error) {}
				console.error(errorText)
				this.priorityPush(this.msgbuilder(
					`RER00;${trimmed};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`,
					data
				))
			})
			return
		}
		if (data.indexOf(">SXATT01") != -1) {
			this.priorityPush(this.msgbuilder("RXATT01", data));
			this.trips_events = true;
			storage.set(`${this.name}_trips_events`, true);
			return;
		}
		if (data.indexOf(">SXATTU0") != -1) {
			this.priorityPush(this.msgbuilder("RXATTU0", data));
			this.trips_events = false;
			storage.set(`${this.name}_trips_events`, false);
			return;
		}
		// OUTPUTS
		if (data.indexOf(">SSSXP") != -1) {
			let output = parseInt(data[6]);
			let status = parseInt(data[7]);
			console.info(`OUT${output} changed to ${status}`);
			this.onCommand({
				command: "set out",
				io: output,
				state: status == 1 ? "on" : "off"
			});
			this.priorityPush(this.msgbuilder(">RSSXP" + output + status, data));
			return;
		}
		if (data.indexOf(">SSSXE") != -1) {
			// Handle this increased 2 because we used expaned outputs for additionals outputs
			let output = parseInt(data[6]);
			let status = parseInt(data[7]);
			console.info(`OUT${output + 2} changed to ${status}`);
			this.onCommand({
				command: "set out",
				io: output + 2,
				state: status == 1 ? "on" : "off"
			});
			this.priorityPush(this.msgbuilder(">RSSXE" + output + status, data));
			return;
		}
		// SEND EVENT
		if (data.indexOf(">SXACT0") != -1) {
			let evtxt = data.substring(data.indexOf(">SXACT0") + 7, data.indexOf(">SXACT0") + 9);
			this.priorityPush(this.msgbuilder(">RXACT0" + evtxt, data));
			this.onCommand({
				command: "send event",
				code: evtxt,
				label: "trckpnt"
			});
			return;
		}
		// COUNTERS
		if (data.indexOf(">SXAVOV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("VO", newval);
			let msg = this.msgbuilder("RXAVOV" + newval, data);
			this.priorityPush(msg);
			return;
		}
		if (data.indexOf(">SXACLV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CL", newval);
			let msg = this.msgbuilder("RXACLV" + newval, data);
			this.priorityPush(msg);
			return;
		}
		if (data.indexOf(">SXACEV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CE", newval);
			let msg = this.msgbuilder("RXACEV" + newval, data);
			this.priorityPush(msg);
			return;
		}
		if (data.indexOf(">SXACSV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CS", newval);
			let msg = this.msgbuilder("RXACSV" + newval, data);
			this.priorityPush(msg);
			return;
		}
		if (data.indexOf(">SXACRV") != -1) {
			let newval = pad(Number(data.substring(7, data.indexOf(";", 7))), 9);
			this.adjustCounter("CR", newval);
			let msg = this.msgbuilder("RXACRV" + newval, data);
			this.priorityPush(msg);
			return;
		}
		// Capture serial photo
		if (data.indexOf(">SXASC") != -1) {
			if (data.indexOf(">SXASCM0") != -1) {
				this.ftp = {}
				storage.remove(`${this.name}_ftp_config`);
				this.priorityPush(this.msgbuilder("RXASCM0", data));
			}
			if (data.indexOf(">SXASCM202") != -1){
				// Activate camera Action: Prepare Yourself
				this.priorityPush(this.msgbuilder(">RXASCM202", data));
			}
			if (data.indexOf(">SXASCP99") != -1) {
				Utils.OSExecute('apx-serial fatigue_sensor capture_photo')
				.then(response => {
					this.priorityPush(this.msgbuilder(">RXASCP99", data));
				})
				.catch(err => {
					let code = err.error ? err.error.code : "1";
					let errorText = JSON.stringify(err);
					this.priorityPush(this.msgbuilder(
						`RXAEC::${command};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`, data)
					);
				})
			}
			return;
		}
		if (data.indexOf(">SSSU051") != -1) {
			Utils.OSExecute("apx-serial fatigue_sensor capture_photo")
			this.priorityPush(this.msgbuilder(">RSSU051", data));
			return;
		}
		// Position
		if (data.indexOf(">SSSU061") != -1) {
			this.priorityPush(this.msgbuilder(">RSSU061", data));
			this.onCommand({
				command: "send event",
				code: "19",
				label: "position"
			});
			return;
		}
		// MDT
		if (data.indexOf(">STX") != -1) {
			let mdt = data.split(">STX")[1].split(";")[0]
			Utils.OSExecute(`apx-serial mdt send '${mdt}'`)
			.then(response =>{
				this.priorityPush(this.msgbuilder(`RTX${mdt}`, data));
			})
			.catch(error =>{
				console.error(error)
				this.priorityPush(this.msgbuilder(`RER00;${trimmed}`, data))
			})
			return;
		}
		// FTP
		if (data.indexOf(">SXAFSA1") != -1) {
			//>SXAFSA1dev.pegasusgateway.com:7021;ID=450000210622036;SI=5-28bfc<
			let split = data
				.split(">SXAFSA1")[1]
				.split(";")[0]
				.split(":");
			this.ftp.server = split[0];
			this.ftp.port = split[1];
			this.priorityPush(this.msgbuilder("RXAFSA1" + this.ftp.server + ":" + this.ftp.port, data));
			storage.hset(`${this.name}_ftp_config`, 'server', split[0]);
			storage.hset(`${this.name}_ftp_config`, 'port', split[1]);
			return
		}
		if (data.indexOf(">SXAFSL") != -1) {
			this.ftp.username = data.split(">SXAFSL")[1].split(";")[0];
			this.priorityPush(this.msgbuilder("RXAFSL" + this.ftp.username, data));
			storage.hset(`${this.name}_ftp_config`, 'username', this.ftp.username);
			return
		}
		if (data.indexOf(">SXAFSP") != -1) {
			this.ftp.password = data.split(">SXAFSP")[1].split(";")[0];
			this.priorityPush(this.msgbuilder("RXAFSP" + this.ftp.password, data));
			storage.hset(`${this.name}_ftp_config`, 'password', this.ftp.password);
			return
		}
		// SafeImmo Signal
		if (data.indexOf(">SXASES") != -1) {
			let value = data.trim()[7]
			this.onCommand({
				isSyrusLang: true,
				command: `set variable $safe_immo ${value}`
			}).then(response => {
				this.priorityPush(this.msgbuilder(`RXASES${value}`, data))
			}).catch(err => {
				let code = err.error ? err.error.code : "1"
				let errorText = JSON.stringify(err)
				if (err.errorText) {
					errorText = err.errorText
				} else if (err.output) {
					errorText = err.output
				}
				try {
					errorText = JSON.parse(errorText).error
				} catch (error) { }
				console.error(errorText)
				this.priorityPush(this.msgbuilder(
					`RER00;${trimmed};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`,
					data
				))
			})
			return
		}
		// MISC
		if (data.indexOf(">SXABR9600") != -1) {
			this.priorityPush(this.msgbuilder("RXABR9600", data));
			return;
		}
		if (data.indexOf(">SXAIBC") != -1) {
			// this.events.publish("unlink", {});
			this.priorityPush(this.msgbuilder("RXAIBC", data));
			return;
		}
		if (data.indexOf(">SRT;SFBUFF") != -1) {
			this.queue.clear();
			this.priorityPush(this.msgbuilder("RRT;SFBUFF", data));
			return;
		}
		if (data.indexOf(">SXAEC::") != -1) {
			let command = data
				.replace(">SXAEC::", "")
				.split("<")[0]
				.split(";ID=")[0]
				.split(";SI=")[0];
			console.log("executed remote command", command);
			Utils.OSExecute(command)
				.then(response => {
					if (!response) response = "OK";
					this.priorityPush(
						this.msgbuilder(
							`RXAEC::${command.replace(/\n/g, "")};0;${
								typeof response == "string"
									? response.substring(0, 512).replace(/\n/g, "")
									: JSON.stringify(response).substring(0, 512).replace(/\n/g, "")
							}`,
							data
						)
					);
				})
				.catch(err => {
					let code = err.error ? err.error.code : "1";
					let errorText = JSON.stringify(err);
					if (err.errorText) {
						errorText = err.errorText;
					} else if (err.output) {
						errorText = err.output;
					}
					try {
						let newErr = JSON.parse(errorText).error;
						if (newErr) errorText = `${newErr}`;
					} catch (error) {}
					this.priorityPush(this.msgbuilder(
						`RXAEC::${command};${code};${errorText.replace(/\n/g, "").substring(0, 512)}`, data)
					);
				});
			return;
		}

		/**
		 * Garmin Mode Implementation
		 */
		if (data.indexOf(">QXAGM") != -1 || data.indexOf(">SXAGM") != -1) {
			this.garminHandler(data);
			return;
		}

		// Catch all unhadled messages
		this.priorityPush(this.msgbuilder(`RER00;${trimmed}`, data))
	}

	/**
	 * this function is intended for handling data received for Garmin plugin from syrus3, not used right now but keep it here for reference
	 * @param {string} data the taip data
	 */
	garminHandler(data) {
		// the gateway ask for the status of garmin
		if (data.indexOf(">QXAGMI") != -1) {
			this.priorityPush(this.msgbuilder(`>RXAGMI${this.imei.substring(5)}00001${pad("00570", 5)}0000`, data));
			return;
		}

		// The Gateway ask if the garmin device is connected
		if (data.indexOf(">SXAGME0") != -1) {
			this.priorityPush(this.msgbuilder(">RXAGME0", data));
			return;
		}

		/** The gateway send a text message */
		if (data.indexOf(">SXAGMTS") != -1) {
			this.priorityPush(this.msgbuilder(data.split(";")[0].replace(">SXAGM", ">RXAGM"), data));
			// let id = parseInt(data.split(";")[0].substring(8, 13));
			// let timestamp = data.split(";")[0].substring(13, 24);
			// let msg = data
			// 	.split(";")[0]
			// 	.substring(24)
			// 	.autoLink({ target: "_blank", rel: "nofollow" });
		}

		// Cath all unhadled command from Garmin mode
		if (data.indexOf(">SXAGM") != -1) {
			this.priorityPush(this.msgbuilder(data.split(";")[0].replace(">SXAGM", ">RXAGM"), data));
			return;
		}
	}

	/**
	 * Converts Json Data from the events in a TAIP event text
	 *
	 * @param {*} data Json Data received from the syrusjs variables and fieldsets
	 * @returns TAIP text message
	 *
	 * @memberOf Parser
	 */
	async JSONtoTAIP(data, opts, fields, label = "trckpnt", code = "01") {
		this.last_data = data;
		let utc_now = dayjs.utc()
		/*
		we will use gnss.timestamp from the backlog when available,
		if not we use the system clock when getting signals,
		this is pending a fix from the gps core
		 */
		if (data['$$backlog']){
			utc_now = dayjs.utc(data['$gnss.timestamp']*1e3)
		}
		let weeks = -GPS_EPOCH.diff(utc_now, "weeks")
		let day = utc_now.day()
		let seconds = pad(utc_now.diff(utc_now.startOf("day"), "seconds"), 5)
		let lat = getLat(data["$gnss.latitude"] || 0)
		let long = getLon(data["$gnss.longitude"] || 0)
		let speed = data["$gnss.speed"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.speed"] * 2.23694)), 3) : "000"
		let heading = data["$gnss.bearing"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.bearing"])), 3) : "000"
		let k = data["$gnss.fix"] || "0"
		let l = ""
		if (!data["$gnss.timestamp"]) l = "0"
		else l = dayjs(data["$gnss.timestamp"]).diff(utc_now, "seconds") < 10 ? "2" : "1"
		let altitude = data["$gnss.altitude"] ? ";AL=" + Math.floor(data["$gnss.altitude"]) : ""
		let dop = ""
		if (data["$gnss.hdop"]) {
			dop = `;DOP=${parseInt(data["$gnss.pdop"] * 100)},${parseInt(data["$gnss.hdop"] * 100)},${parseInt(data["$gnss.vdop"] * 100)}`
		}
		let bl = "";
		if (data["$battery.mv"] && Math.floor(data["$battery.mv"]) > -1) bl = ";BL=" + Math.floor(data["$battery.mv"]);
		let evLabel = ";EVLBL=" + label;
		let ix = "";
		let adaw = "";
		if(fields["$io"]){
			ix = getIO(data);
			adaw = getAnalogs(data);
		}
		let sc = await getSC(fields, data);
		let satcom = await getSatcom(fields, data);
		let imeipart = ";ID=" + this.imei;
		let ackpart = ""
		if (opts.ack == "disabled"){
			ackpart = `;SA=0,${this.ack_counter}`
		}
		if (opts.ack == "seq"){
			ackpart = `;SA=1,${this.ack_counter}`
			await storage.set(`${this.name}_ack`, ++this.ack_counter);
		}
		let aidpart = "";
		let sv = `;SV=${data["$gnss.satsActive"] || 0}`;
		// GPS Acceleration
		let ac = `;AC=${data["$gnss.acceleration.value"] || 0 }`;
		code = pad(code, 2);
		let ecu = "";
		if(fields["$ecu"]){
			ecu = ECUToTAIP(data);
		}
		let counters = getCounters(fields, data);
		let ib_is = getIB_IS(fields, data)
		let ea = getEA(fields, data)
		// RFID
		let ri = getRI(fields, data)
		
		// Techonoton fuel
		let ff = getFF(fields, this.last_data)

		let taip = ""

		// photo
		let media = await this.getPS(fields, data, opts)
		// MDT
		if (opts.trigger && opts.trigger.includes('@mdt.message')){
			let mdt = data['$mdt.message']
			taip += `>RTX${mdt};EV`
		}
		if (!taip.length) taip += `>REV`
		taip += `${code}${weeks}${day}${seconds}${lat}${long}${speed}${heading}${k}${l}${altitude}`
		taip += `${evLabel}${dop}${bl}${ix}${adaw}${sv}${ac}`
		taip += `${counters}${ib_is}`
		taip += `${ea}${aidpart}${sc}`
		taip += `${ri}`
		taip += `${ff}`
		taip += `${ecu}`
		taip += `${media}`
		taip += `${satcom}`
		taip += `${ackpart}${imeipart}<`;
		return taip;
	}

	/**
	 * RPV taip constructor
	 * @param {string} data the variables from syrusjs
	 */
	RPV(data = {}) {
		const utc_now = dayjs.utc()
		let seconds = pad(utc_now.diff(utc_now.startOf("day"), "seconds"), 5);
		let lat = getLat(data["$gnss.latitude"] || 0);
		let long = getLon(data["$gnss.longitude"] || 0);
		let speed = data["$gnss.speed"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.speed"] * 2.23694)), 3) : "000";
		let heading = data["$gnss.bearing"] >= 0 ? pad(Math.round(Math.abs(data["$gnss.bearing"])), 3) : "000";
		let k = data["$gnss.fix"] || 0;
		let l
		if (!data["$gnss.timestamp"]) l = 0;
		else l = dayjs(data["$gnss.timestamp"]).diff(utc_now, "seconds") < 10 ? "2" : "1";
		return `>RPV${seconds}${lat}${long}${speed}${heading}${k}${l}`;
	}

	/**
	 * helper to construct response to the gateway, returns the TAIP command with ID, SI
	 *
	 * @param {string} message string that will be send to gateway without ID or SI parts
	 * @param {string} original the original message received from gateway
	 * @returns {string} the TAIP text command to send to gateway
	 *
	 */
	msgbuilder(message, original) {
		if (!original) original = ''

		let text = message.trim()

		if (!text.startsWith(">")) text = `>${text}`

		if (original.includes(";SI=")) {
			text += `;SI=${original.split(';SI=')[1].split('<')[0].split(';')[0]}`
		}

		if (!text.includes(";ID=")) {
			text += `;ID=${this.imei}`
		}

		if (!text.endsWith("<")) text = `${text}<`

		return text;
	}

	/**
	 * allow to adjust counters from the fieldset defined
	 * @param {String} key the value to adjust
	 * @param {Number} newValue the new value to set
	 */
	async adjustCounter(key, newValue) {
		const DEFAULTS_COUNTERS = {
			VO: "$counters.globals.odometer",
			CL: "$counters.globals.idle_time",
			CE: "$counters.globals.ignition_time",
			CS: "$counters.globals.over_speed",
			CR: "$counters.globals.over_rpm"
		};
		let definition = null
		if (this.fields) definition = this.fields[key];
		if (!definition) definition = DEFAULTS_COUNTERS[key];
		let [signal, namespace, value] = definition.split(".");
		if (signal !== "$counters") console.error(`${key} is not assigned to an apx counter`);
		let promise = Utils.OSExecute(`apx-counter set ${namespace} ${value.toUpperCase()} ${Math.round(Number(newValue))}`);
		promise.catch(console.error);
		return promise;
	}
}

export const TAIP = MediaManagerMixin(BacklogMixin(BaseTAIP))
